/*****************************************************************************
 *****************************************************************************
 Copyright (c) 2000-2001, AEP Ltd. 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.

 3. Neither the name of AEP Ltd. nor the names of its contributors 
 may be used to endorse or promote products derived from this software 
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

 *****************************************************************************
****************************************************************************/

/**********************************************************************
*                                                                     *
* AEP Ltd.                                                            *
*                                                                     *
* This software is supplied under the terms of the license included   *
* above.  All use of this software must be in accordance with the     *
* terms of that license.                                              *
*                                                                     *
**********************************************************************/

/*****************#defines********************/

/*Settings for calculating the rate of mod exps. 
  Need to keep the number of threads reasonably high to get good performance, and need to 
  keep the number of mod exps fairly high to average out the time lost starting up the threads 
  over a number of calls. However, we also need to keep the time the user is waiting to a minimum
  (while not compromising performance).
  A value of 150 for both of these seems to work well (when logging is turned off)*/
#define NUM_MOD_EXPS 150
#define NUMTHREADS 150


/*How many checks to run on the random numbers*/
#define NUM_HARDWARE_RAND_TESTS 1
#define NUM_SOFTWARE_RAND_TESTS 10

/*Note : for the randomness test we need 2500 bytes of random numbers. We can't generate
  this all in one go, so need to break it up into chunks and generate each chunk seperately.
  Here we define the size of the chunks*/
#define HARDWARE_RANDOM_CHUNKS 50

/*This is used to limit the number of errors you get when calculating the rate of Mod Exps.
  If you exceed this number of errors the threads will stop*/
#define ERROR_LIMIT 50

#define HARDWARE_RANDOM 1
#define PSEUDO_RANDOM 2
#define VERBOSE_OPTION "-v"


/*Error codes*/
/*An error code will be returned from the program to show what the results were, it may be one of the regular
 AEP error codes or one of these*/
#define HARDWARE_RAND_INCORRECT 0x11000000
#define PSEUDO_RAND_INCORRECT 0x12000000
#define AEP_MOD_EXP_INCORRECT 0x13000000
#define AEP_MOD_EXP_CRT_INCORRECT 0x14000000
#define SOFTWARE_MOD_EXP_INCORRECT 0x15000000
#define EXCEEDED_MOD_EXP_RATE_ERROR_LIMIT 0x16000000



/*************Function prototypes*************/

/*Function definition for the Mod Exp thread*/
static void *ptrModExpThread();
static void count_exits();

/*Callback functions*/
static AEP_RV GetBigNumSize(void* ArbBigNum, AEP_U32* BigNumSize);
static AEP_RV MakeAEPBigNum(void* ArbBigNum, AEP_U32 BigNumSize, unsigned char* AEP_BigNum);
static AEP_RV ConvertAEPBigNum(void* ArbBigNum, AEP_U32 BigNumSize, unsigned char* AEP_BigNum);

/*Function to print error message*/
static void printErrorMessage(AEP_RV returnVal);

/*****************Global Data*****************/
int modExpErrorCount;
int verboseMode = 0;

pthread_cond_t  ThreadsFinished;
pthread_mutex_t count_lock;
pthread_attr_t attr1;

AEP_CONNECTION_HNDL hConnection[1000];

int finished;
