/*****************************************************************************
 *****************************************************************************
 Copyright (c) 2000-2001, AEP Ltd. 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.

 3. Neither the name of AEP Ltd. nor the names of its contributors 
 may be used to endorse or promote products derived from this software 
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

 *****************************************************************************
****************************************************************************/

/**********************************************************************
*                                                                     *
* AEP Ltd.                                                            *
*                                                                     *
* This software is supplied under the terms of the license included   *
* above.  All use of this software must be in accordance with the     *
* terms of that license.                                              *
*                                                                     *
**********************************************************************/
/****************************************************************************************************
                                                                Montgomery.c
                                                           Barbara Hamill
*****************************************************************************************************/

#include <stdlib.h>
#include <string.h> 
#include <time.h>
#include <stdio.h>
#include "aepmalloc.h"
#include "montgomery.h"

/*Global Variables to keep track of number of mallocs and frees used*/
int mallocCount = 1;
int mallocNew = 1;
int mallocSet = 1;
int mallocSetDiff = 1;
int mallocExpand = 1;
int mallocCTXNew = 1;
int freeExpand = 1;
int freeCount = 1;


int incBase =0 ;
int incPower = 0;
int incMod = 0;


/*****************************************************************************************************/
	/*Create a new Big Number - Code from OpenSSL*/
	BIGNUM *BN_new(void)
	{
		BIGNUM *ret;

		if ((ret=(BIGNUM *)AEP_malloc(sizeof(BIGNUM))) == NULL)
		{
				printf("\nAEP_malloc Failure\n");
				return(NULL);
		}
		mallocCount++;
		mallocNew++;

		ret->flags=BN_FLG_MALLOCED;
		ret->top=0;
		ret->neg=0;
		ret->max=0;
		ret->d=NULL;
    
		return(ret);
	}

/************************************************************************************************/
	/*Initialise Structure entries to zero - Code from OpenSSL*/
	void BN_init(BIGNUM *a)
	{
		memset(a,0,sizeof(BIGNUM));
		a->flags = BN_FLG_MALLOCED;
	}

/************************************************************************************************/
/*Sets a bignumber to the hardcoded number below - For testing purposes only*/
BIGNUM* BN_SetBase()
{       
        
    BIGNUM *a;
    BN_ULONG lsb = 0x00000001L;
    a = BN_new();
    BN_init(a);
    a->d = (BN_ULONG *)AEP_malloc(sizeof(BN_ULONG)*32);     /*Allocate 8 blocks of memory*/

    a->top          =32;
    a->max          =32;            /*4 Blocks*/
    
    
    
                    
    
    /*Each entry is 32 bits*/
/*      a->d[0] = 4;*/

    a->d[0] = lsb + (incBase * 2);
    incBase++;
    
    if(incBase == 66)
            incBase = 0;

    a->d[0] = 0x5CBFF264uL;
    a->d[1] = 0x39266AEAuL;  
    a->d[2] = 0x3CD4E608uL;  
    a->d[3] = 0x9D9B99B7uL;
    a->d[4] = 0x05000410uL;
    a->d[5] = 0xF70D0205uL;
    a->d[6] = 0x2A864886uL;
    a->d[7] = 0x300C0608uL;
    a->d[8] = 0xFF003020uL;
    a->d[9] = 0xffffffffuL;
    a->d[10] = 0xffffffffuL;
    a->d[11] = 0xffffffffuL;
    a->d[12] = 0xffffffffuL;
    a->d[13] = 0xffffffffuL;
    a->d[14] = 0xffffffffuL;
    a->d[15] = 0xffffffffuL;
    
    a->d[16] = 0x0uL;        
    a->d[17] = 0x0uL;
    a->d[18] = 0x0uL;
    a->d[19] = 0x0uL;
    a->d[20] = 0x0uL;
    a->d[21] = 0x0uL;
    a->d[22] = 0x0uL;
    a->d[23] = 0x0uL;
    
    a->d[24] = 0x0uL;        
    a->d[25] = 0x0uL;
    a->d[26] = 0x0uL;
    a->d[27] = 0x0uL;
    a->d[28] = 0x0uL;
    a->d[29] = 0x0uL;
    a->d[30] = 0x0uL;
    a->d[31] = 0xffffffffuL;
/*      
    a->d[32] = 0x5CBFF264uL;
    a->d[33] = 0x39266AEAuL; 
    a->d[34] = 0x3CD4E608uL; 
    a->d[35] = 0x9D9B99B7uL;
    a->d[36] = 0x05000410uL;
    a->d[37] = 0xF70D0205uL;
    a->d[38] = 0x2A864886uL;
    a->d[39] = 0x300C0608uL;
    a->d[40] = 0xFF003020uL;
    a->d[41] = 0xffffffffuL;
    a->d[42] = 0xffffffffuL;
    a->d[43] = 0xffffffffuL;
    a->d[44] = 0xffffffffuL;
    a->d[45] = 0xffffffffuL;
    a->d[46] = 0xffffffffuL;
    a->d[47] = 0xffffffffuL;
    
    a->d[48] = 0x0uL;        
    a->d[49] = 0x0uL;
    a->d[50] = 0x0uL;
    a->d[51] = 0x0uL;
    a->d[52] = 0x0uL;
    a->d[53] = 0x0uL;
    a->d[54] = 0x0uL;
    a->d[55] = 0x0uL;
    
    a->d[56] = 0x0uL;        
    a->d[57] = 0x0uL;
    a->d[58] = 0x0uL;
    a->d[59] = 0x0uL;
    a->d[60] = 0x0uL;
    a->d[61] = 0x0uL;
    a->d[62] = 0x1uL;
    a->d[63] = 0x3fffffffuL;
    a->d[64] = 0x5CBFF264uL;
    a->d[65] = 0x39266AEAuL; 
    a->d[66] = 0x3CD4E608uL; 
    a->d[67] = 0x9D9B99B7uL;
    a->d[68] = 0x05000410uL;
    a->d[69] = 0xF70D0205uL;
    a->d[70] = 0x2A864886uL;
    a->d[71] = 0x300C0608uL;
    a->d[72] = 0xFF003020uL;
    a->d[73] = 0xffffffffuL;
    a->d[74] = 0xffffffffuL;
    a->d[75] = 0xffffffffuL;
    a->d[76] = 0xffffffffuL;
    a->d[77] = 0xffffffffuL;
    a->d[78] = 0xffffffffuL;
    a->d[79] = 0xffffffffuL;
    
    a->d[80] = 0x0uL;        
    a->d[81] = 0x0uL;
    a->d[82] = 0x0uL;
    a->d[83] = 0x0uL;
    a->d[84] = 0x0uL;
    a->d[85] = 0x0uL;
    a->d[86] = 0x0uL;
    a->d[87] = 0x0uL;
    
    a->d[88] = 0x0uL;        
    a->d[89] = 0x0uL;
    a->d[90] = 0x0uL;
    a->d[91] = 0x0uL;
    a->d[92] = 0x0uL;
    a->d[93] = 0x0uL;
    a->d[94] = 0x0uL;
    a->d[95] = 0xffffffffuL;
    
    a->d[96] = 0x5CBFF264uL;
    a->d[97] = 0x39266AEAuL; 
    a->d[98] = 0x3CD4E608uL; 
    a->d[99] = 0x9D9B99B7uL;
    a->d[100] = 0x05000410uL;
    a->d[101] = 0xF70D0205uL;
    a->d[102] = 0x2A864886uL;
    a->d[103] = 0x300C0608uL;
    a->d[104] = 0xFF003020uL;
    a->d[105] = 0xffffffffuL;
    a->d[106] = 0xffffffffuL;
    a->d[107] = 0xffffffffuL;
    a->d[108] = 0xffffffffuL;
    a->d[109] = 0xffffffffuL;
    a->d[110] = 0xffffffffuL;
    a->d[111] = 0xffffffffuL;
    
    a->d[112] = 0x0uL;       
    a->d[113] = 0x0uL;
    a->d[114] = 0x0uL;
    a->d[115] = 0x0uL;
    a->d[116] = 0x0uL;
    a->d[117] = 0x0uL;
    a->d[118] = 0x0uL;
    a->d[119] = 0x0uL;
    
    a->d[120] = 0x0uL;       
    a->d[121] = 0x0uL;
    a->d[122] = 0x0uL;
    a->d[123] = 0x0uL;
    a->d[124] = 0x0uL;
    a->d[125] = 0x0uL;
    a->d[126] = 0x1uL;
    a->d[127] = 0x3fffffffuL;
*/

    
    return a;
}


BIGNUM* BN_SetPower()
{       
        
    BIGNUM *a;
    BN_ULONG lsb = 0x00000001;
    a = BN_new();
    BN_init(a);
    a->d = (BN_ULONG *)AEP_malloc(sizeof(BN_ULONG)*16);     /*Allocate 8 blocks of memory*/
    
    a->top          = 16;
    a->max          = 16;           /*4 Blocks*/
    
    
    /*a->d[0] = 3;*/
    a->d[0] = lsb + (incPower * 2);
    incPower++;
    
    if(incPower == 66)
            incPower = 0;


    
    a->d[0] = 0xFFFFFFFFuL;  
    a->d[1] = 0x12345678uL;  
    a->d[2] = 0x87654321uL;
    a->d[3] = 0xFFFFFFFFuL;
    a->d[4] = 0xFFFFFFFFuL;
    a->d[5] = 0x0uL;
    a->d[6] = 0x0uL;
    a->d[7] = 0x0uL;
    a->d[8] = 0x0uL;
    a->d[9] = 0x0uL;
    a->d[10] = 0x0uL;
    a->d[11] = 0x0uL;
    a->d[12] = 0x0uL;
    a->d[13] = 0x0uL;
    a->d[14] = 0x0uL;
    a->d[15] = 0xfuL;
/*              
    a->d[16] = 0x0uL;        
    a->d[17] = 0x0uL;
    a->d[18] = 0x0uL;
    a->d[19] = 0x0uL;
    a->d[20] = 0x0uL;
    a->d[21] = 0x0uL;
    a->d[22] = 0x0uL;
    a->d[23] = 0x0uL;
    a->d[24] = 0x0uL;
    a->d[25] = 0x0uL;
    a->d[26] = 0x0uL;
    a->d[27] = 0x0uL;
    a->d[28] = 0x0uL;
    a->d[29] = 0x0uL;
    a->d[30] = 0xf0000001uL;
    a->d[31] = 0xffffffffuL;
    a->d[32] = 0xFFFFFFFFuL; 
    a->d[33] = 0x12345678uL; 
    a->d[34] = 0x87654321uL;
    a->d[35] = 0xFFFFFFFFuL;
    a->d[36] = 0xFFFFFFFFuL;
    a->d[37] = 0x0uL;
    a->d[38] = 0x0uL;
    a->d[39] = 0x0uL;
    a->d[40] = 0x0uL;
    a->d[41] = 0x0uL;
    a->d[42] = 0x0uL;
    a->d[43] = 0x0uL;
    a->d[44] = 0x0uL;
    a->d[45] = 0x0uL;
    a->d[46] = 0x0uL;
    a->d[47] = 0x0uL;
    
    a->d[48] = 0x0uL;        
    a->d[49] = 0x0uL;
    a->d[50] = 0x0uL;
    a->d[51] = 0x0uL;
    a->d[52] = 0x0uL;
    a->d[53] = 0x0uL;
    a->d[54] = 0x0uL;
    a->d[55] = 0x0uL;
    a->d[56] = 0x0uL;
    a->d[57] = 0x0uL;
    a->d[58] = 0x0uL;
    a->d[59] = 0x0uL;
    a->d[60] = 0x0uL;
    a->d[61] = 0x0uL;
    a->d[62] = 0xf0000001uL;
    a->d[63] = 0xffffffffuL;
*/

    return a;
}
/************************************************************************************************/
/*Sets a bignumber to the hardcoded number below - For testing purposes only*/
BIGNUM* BN_SetMod()
{
    BIGNUM* a;
    BN_ULONG lsb = 0x00000001L;
    
    a = BN_new();
    BN_init(a);

    a->d = (BN_ULONG *)AEP_malloc(sizeof(BN_ULONG)*16);     /*Allocate 8 blocks of memory*/
    
    a->top          = 16;
    a->max          = 16;           /*4 Blocks*/
    /*Each entry is 32 bits*/
    
    
    a->d[0] = lsb + (incMod * 2);
    incMod++;
    
    if(incMod == 66)
            incMod = 0;

    a->d[0] = 0x5CBFF265uL;
    a->d[1] = 0x39266AEAuL;  
    a->d[2] = 0x3CD4E608uL;  
    a->d[3] = 0x9D9B99B7uL;
    a->d[4] = 0x05000410uL;
    a->d[5] = 0xF70D0205uL;
    a->d[6] = 0x2A864886uL;
    a->d[7] = 0x300C0608uL;
    a->d[8] = 0xFF003020uL;
    a->d[9] = 0xffffffffuL;
    a->d[10] = 0xffffffffuL;
    a->d[11] = 0xffffffffuL;
    a->d[12] = 0xffffffffuL;
    a->d[13] = 0xffffffffuL;
    a->d[14] = 0xffffffffuL;
    a->d[15] = 0xffffffffuL;
/*              
    a->d[16] = 0x0uL;        
    a->d[17] = 0x0uL;
    a->d[18] = 0x0uL;
    a->d[19] = 0x0uL;
    a->d[20] = 0x0uL;
    a->d[21] = 0x0uL;
    a->d[22] = 0x0uL;
    a->d[23] = 0x0uL;
    
    a->d[24] = 0x0uL;        
    a->d[25] = 0x0uL;
    a->d[26] = 0x0uL;
    a->d[27] = 0x0uL;
    a->d[28] = 0x0uL;
    a->d[29] = 0x0uL;
    a->d[30] = 0x0uL;
    a->d[31] = 0x3fffffffLuL;
    a->d[32] = 0xFFFFFFFFLuL; 
    a->d[33] = 0x12345678LuL; 
    a->d[34] = 0x87654321LuL;
    a->d[35] = 0xFFFFFFFFLuL;
    a->d[36] = 0xFFFFFFFFLuL;
    a->d[37] = 0x0uL;
    a->d[38] = 0x0uL;
    a->d[39] = 0x0uL;
    a->d[40] = 0x0uL;
    a->d[41] = 0x0uL;
    a->d[42] = 0x0uL;
    a->d[43] = 0x0uL;
    a->d[44] = 0x0uL;
    a->d[45] = 0x0uL;
    a->d[46] = 0x0uL;
    a->d[47] = 0x0uL;
    
    a->d[48] = 0x0uL;        
    a->d[49] = 0x0uL;
    a->d[50] = 0x0uL;
    a->d[51] = 0x0uL;
    a->d[52] = 0x0uL;
    a->d[53] = 0x0uL;
    a->d[54] = 0x0uL;
    a->d[55] = 0x0uL;
    a->d[56] = 0x0uL;
    a->d[57] = 0x0uL;
    a->d[58] = 0x0uL;
    a->d[59] = 0x0uL;
    a->d[60] = 0x0uL;
    a->d[61] = 0x0uL;
    a->d[62] = 0x00005001uL;
    a->d[63] = 0x3fffffffuL;
*/

	return a;
}

#if 0
/*      THIS FUNCTION IS NOT USED AT THE PRESENT!!! */
/************************************************************************************************/
/*Prints the hex value of a big number*/
void PrintBigNum(BIGNUM *a)
{
        
    BN_ULONG mask= (unsigned long)0xF0000000;                               /*mask byte*/
    int i;
    
    if(a->top == 0)
            printf("00000000");

    /*If the number is negative, preceed it with a minus sign*/
    if(a->neg == 1)
            printf("-");
    
    for(i=((a->top)-1);i>=0;i--)                    /* 'for' each chunk */
    {
        if(a->d[i])
        {
            while(!(a->d[i] & mask) )               /*insert leading zeros*/
            {
                    printf("0");
                    mask = mask>>4;                         /*shift to the next least significant bit*/
            }
            mask= 0xF0000000;                               /*Reset the mask byte*/
            printf("%X",a->d[i] );          
        }
        else
            printf("00000000");

    }
    printf("\n");
}
#endif

/************************************************************************************************/
/*Set the big number to a particular word - Code from OpenSSL */
int BN_set_word(BIGNUM *a, BN_ULONG w)
{
    
    if (bn_expand(a,sizeof(BN_ULONG)) == NULL) return(0);
    
    a->top=0;
    a->d[0]=(BN_ULONG)w &BN_MASK2;
    if (a->d[0] != 0) a->top=1;
        
	return(1);
}
        
/************************************************************************************************/
/*Copy b into a - Code from OpenSSL*/
BIGNUM *BN_copy(BIGNUM *a, BIGNUM *b)
{
    int i;
    BN_ULONG *A;
    const BN_ULONG *B;

    bn_check_top(b);
    if (a == b) return(a);
    /*make sure a has enough words to hold the copy of b*/
    if (bn_wexpand(a,b->top) == NULL) return(NULL);

    A=a->d;
    B=b->d;
    for(i = 0; i < b->top; i++)
            a->d[i] = b->d[i];

    a->top=b->top;
    a->neg = b->neg;
    if ((a->top == 0) && (a->d != NULL))
            a->d[0]=0;
    return(a);
}
/************************************************************************************************/
/*Expand the memory allocated to b to the number of words specified - Code from OpenSSL*/
BIGNUM *bn_expand2(BIGNUM *b, int words)
{
    BN_ULONG *A;
    const BN_ULONG *B;
    int i;

    bn_check_top(b);

    if (words > b->max)
    {
        A=(BN_ULONG *)AEP_malloc(sizeof(BN_ULONG)*(words+1));
        mallocExpand++;
        
        mallocCount++;
        if (A == NULL)
        {
            printf("\nMalloc Failure\n");
            return(NULL);
        }
        B=b->d;
        for(i = 0; i < words+1; i++)
            A[i] = 0;

        for(i = 0; i < b->top; i++)
            A[i] = b->d[i];
        
        if(B != NULL)
        {
            Free(b->d);
            freeCount++;    
            freeExpand++;
        }
        
        b->d=A;
        b->max=words;

    }

    return(b);
}


/************************************************************************************************/
/*Free the memory used by a big number - Code from OpenSSL*/
void BN_clear_free(BIGNUM *a)
{
    int i;

    if (a == NULL) return;
    if (a->d != NULL)
    {
        memset(a->d,0,a->max*sizeof(a->d[0]));
        if (!(BN_get_flags(a,BN_FLG_STATIC_DATA)))
        {
            Free(a->d);
            
            freeCount++;    
        }
    }
    i=BN_get_flags(a,BN_FLG_MALLOCED);
    memset(a,0,sizeof(BIGNUM));
    if (i)
    {
        Free(a);
        
        freeCount++;    
    }
}

/************************************************************************************************/
BN_CTX *BN_CTX_new(void)
{
    BN_CTX *ret;

    ret=(BN_CTX *)AEP_malloc(sizeof(BN_CTX));
    mallocCTXNew++;
    
    mallocCount++;
    if (ret == NULL)
    {
        printf("\nMalloc Failure\n");
        return(NULL);
    }

    BN_CTX_init(ret);
    ret->flags=BN_FLG_MALLOCED;
    return(ret);
}
/************************************************************************************************/
void BN_CTX_init(BN_CTX *ctx)
{
    memset(ctx,0,sizeof(BN_CTX));
    ctx->tos=0;
    ctx->flags=0;

}
/************************************************************************************************/
void BN_CTX_free(BN_CTX *c)
{
    int i;

    if(c == NULL)
        return;

    for (i=0; i <= BN_CTX_NUM; i++)
            BN_clear_free(&(c->bn[i]));
    if (c->flags & BN_FLG_MALLOCED)
    {
        Free(c);
        
        freeCount++;    
    }

}

/**************************************************************************************************************/
/* Add a and b and store the result in r - Code from OpenSSL*/
int BN_add(BIGNUM *r, BIGNUM *a, BIGNUM *b)
{
    BIGNUM *tmp;

    bn_check_top(a);
    bn_check_top(b);

    /*  a +  b      a+b
     *  a + -b      a-b
     * -a +  b      b-a
     * -a + -b      -(a+b)
     */
    if (a->neg ^ b->neg)
    {
        /* only one is negative */
        if (a->neg)
            { tmp=a; a=b; b=tmp; }

        /* we are now a - b */

        if (BN_ucmp(a,b) < 0)
        {
            if (!BN_usub(r,b,a)) return(0);
            r->neg=1;
        }
        else
        {
            if (!BN_usub(r,a,b)) return(0);
            r->neg=0;
        }
        
        return(1);
    }

    if (a->neg) /* both are neg */
        r->neg=1;
    else
        r->neg=0;

    if (!BN_uadd(r,a,b)) return(0);
    
    return(1);
}
/************************************************************************************************/
/* unsigned addition of b to a, r must be large enough - Code from OpenSSL*/
int BN_uadd(BIGNUM *r, BIGNUM *a, BIGNUM *b)
{
    register int i;
    int max,min;
    BN_ULONG *ap,*bp,*rp,carry,t1;
    BIGNUM *tmp;

    
    bn_check_top(a);
    bn_check_top(b);

    if (a->top < b->top)
            { tmp=a; a=b; b=tmp; }
    max=a->top;
    min=b->top;

    if (bn_wexpand(r,max+1) == NULL)
            return(0);

    r->top=max;

    /*for(i = 0; i < b->top; i++)
    {
            add(r->d[i], a->d[i], b->d[i], carry);
    }
    while(i < a->top)
    {
            add(r->d[i], a->d[i], 0, carry);
            i++;
    }
    */


    ap=a->d;
    bp=b->d;
    rp=r->d;
    carry=0;

    carry=bn_add_words(rp,ap,bp,min);
    rp+=min;
    ap+=min;
    bp+=min;
    i=min;

    if (carry)
    {
        while (i < max)
        {
            i++;
            t1= *(ap++);
            if ((*(rp++)=(t1+1)&BN_MASK2) >= t1)
            {
                carry=0;
                break;
            }
        }
        if ((i >= max) && carry)
        {
            *(rp++)=1;
            r->top++;
        }
    }
    if (rp != ap)
    {
        for (; i<max; i++)
                *(rp++)= *(ap++);
    }
    
    return(1);
}

/************************************************************************************************/
/*Adds n words of a and b and stores the result in r - Code from OpenSSL*/
BN_ULONG bn_add_words(BN_ULONG *r, BN_ULONG *a, BN_ULONG *b, int n)
{
    BN_ULONG c,l,t;

    c=0;
    for (;;)
    {
        t=a[0];
        t=(t+c)&BN_MASK2;
        c=(t < c);
        l=(t+b[0])&BN_MASK2;
        c+=(l < t);
        r[0]=l;
        if (--n <= 0) break;

        t=a[1];
        t=(t+c)&BN_MASK2;
        c=(t < c);
        l=(t+b[1])&BN_MASK2;
        c+=(l < t);
        r[1]=l;
        if (--n <= 0) break;

        t=a[2];
        t=(t+c)&BN_MASK2;
        c=(t < c);
        l=(t+b[2])&BN_MASK2;
        c+=(l < t);
        r[2]=l;
        if (--n <= 0) break;

        t=a[3];
        t=(t+c)&BN_MASK2;
        c=(t < c);
        l=(t+b[3])&BN_MASK2;
        c+=(l < t);
        r[3]=l;
        if (--n <= 0) break;

        a+=4;
        b+=4;
        r+=4;
    }

    return((BN_ULONG)c);
}

/************************************************************************************************/      
/*Signed subtraction : r = a - b - Code from OpenSSL*/
int BN_sub(BIGNUM *r, BIGNUM *a, BIGNUM *b)
{
    int max;
    int add=0,neg=0;
    BIGNUM *tmp;


    bn_check_top(a);
    bn_check_top(b);

    if((a->neg == 1) || ( b->neg == 1))
        bn_wexpand(r, a->top+b->top);

    /*  a -  b      a-b
     *  a - -b      a+b
     * -a -  b      -(a+b)
     * -a - -b      b-a
     */
    if (a->neg)
    {
        if (b->neg)
                { tmp=a; a=b; b=tmp; }
        else
                { add=1; neg=1; }
    }
    else
    {
        if (b->neg) { add=1; neg=0; }
    }

    if (add)
    {
        if (!BN_uadd(r,a,b)) return(0);
        
        r->neg=neg;
        return(1);
    }

    /* We are actually doing a - b :-) */

    max=(a->top > b->top)?a->top:b->top;
    if (bn_wexpand(r,max) == NULL) return(0);
    
    if (BN_ucmp(a,b) < 0)
    {
        if (!BN_usub(r,b,a)) return(0);
        r->neg=1;
    }
    else
    {
        if (!BN_usub(r,a,b)) return(0);
        r->neg=0;
    }

    return(1);
}
/************************************************************************************************/
/* unsigned subtraction of b from a, a must be larger than b. */
int BN_usub(BIGNUM *r, BIGNUM *a, BIGNUM *b)
{
    int max,min;
    register BN_ULONG t1,t2,*ap,*bp,*rp;
    int i,carry;

    bn_check_top(a);
    bn_check_top(b);
    

    if (a->top < b->top) /* hmm... should not be happening */
    {
            printf("\nError!  a should be larger than b\n");
            return(0);
    }

    max=a->top;
    min=b->top;
    
    if (bn_wexpand(r,max) == NULL) return(0);

    ap = a->d;
    bp = b->d;
    rp = r->d;


    carry=0;
    for (i=0; i<min; i++)
    {
        t1= *(ap++);
        t2= *(bp++);
        if (carry)
        {
            carry=(t1 <= t2);
            t1=(t1-t2-1)&BN_MASK2;
        }
        else
        {
            carry=(t1 < t2);
            t1=(t1-t2)&BN_MASK2;
        }
        *(rp++)=t1&BN_MASK2;
    }
    if (carry) /* subtracted */
    {
        while (i < max)
        {
            i++;
            t1= *(ap++);
            t2=(t1-1)&BN_MASK2;
            *(rp++)=t2;
            if (t1 > t2) break;
        }
    }
    if (rp != ap)
    {
        for (;;)
        {
            if (i++ >= max) break;
            rp[0]=ap[0];
            if (i++ >= max) break;
            rp[1]=ap[1];
            if (i++ >= max) break;
            rp[2]=ap[2];
            if (i++ >= max) break;
            rp[3]=ap[3];
            rp+=4;
            ap+=4;
        }
    }

    r->top=max;
    bn_fix_top(r);
    return(1);
}
/************************************************************************************************/
/*Shift a to the right, n number of positions and store the result in r - Code from 
  OpenSSL
*/
int BN_rshift(BIGNUM *r, BIGNUM *a, int n)
{
    int i,j,nw,lb,rb;
    BN_ULONG *t,*f;
    BN_ULONG l,tmp;

    nw=n/BN_BITS2;
    rb=n%BN_BITS2;
    lb=BN_BITS2-rb;
    if (nw > a->top)
    {
        BN_zero(r);
        return(1);
    }
    if (r != a)
    {
        if (bn_wexpand(r,a->top-nw+1) == NULL) return(0);
    }

    f= &(a->d[nw]);
    t=r->d;
    j=a->top-nw;
    r->top=j;

    if (rb == 0)
    {
        for (i=j+1; i > 0; i--)
                *(t++)= *(f++);
    }
    else
    {
        l= *(f++);
        for (i=1; i<j; i++)
        {
            tmp =(l>>rb)&BN_MASK2;
            l= *(f++);
            *(t++) =(tmp|(l<<lb))&BN_MASK2;
        }
        *(t++) =(l>>rb)&BN_MASK2;
    }
    *t=0;

            
    bn_fix_top(r);
    return(1);
}

/************************************************************************************************/
        /*compare two unsigned numbers.  If a is the same wordlength but a bigger number than b 
        return 1 else if b is the same wordlength but bigger return -1, if they are the sane
        wordlength and the same number return 0 otherwise return the number of words in the difference
        Code from OpenSSL
        */
        int BN_ucmp(BIGNUM *a, BIGNUM *b)
        {
                int i;
                BN_ULONG t1,t2,*ap,*bp;

                bn_check_top(a);
                bn_check_top(b);

                /*i is difference in word sizes  (a should be bigger than b)*/
                i=a->top-b->top;

                /*if there is a difference return the difference*/
                if (i != 0) return(i);

                /*set pointers to bit chunk arrays*/
                ap=a->d;
                bp=b->d;

                /*set pointer to last index used in array and loop through filled elements*/
                for (i=a->top-1; i>=0; i--)
                {
                        t1= ap[i];
                        t2= bp[i];
                        if (t1 != t2)
                                return(t1 > t2?1:-1);
                }
                return(0);
        }
/************************************************************************************************/

/*
  This function is used to count the exact number of bits in a big number
  Input : Big Number
  Output: The number of bits in this number
        
*/

int countExactNumberOfBits(BIGNUM* a)
{
        
    /*Count the number of bits including leading zeros in the number*/
    int numberOfBits = countBits(a);
    int found = 0;
    int i = 0;
    int j = 0;
    BN_ULONG one = (unsigned long)0x8 << (BN_BITS2 - 4);


    /*Supress leading zeros to get an exact bit count*/
    for(i = a->top-1; i >= 0; i--)
    {

        one = (unsigned long)0x8 << (BN_BITS2 - 4);

        for(j = 0; j < BN_BITS2; j++)
        {
            /*Decrement number of bits until first word is found*/
            if((!(a->d[i] & one)) && (found != 1))
            {
                numberOfBits --;
                one >>= 1;
            }
            else
            {
                found = 1;
                break;
            }
        }
    }

    return numberOfBits;

}
/*****************************************************************************************************/
int countExactNumberOfBitsInWord(BN_ULONG a)
{
    
    /*Count the number of bits including leading zeros in the number*/
    int numberOfBits = BN_BITS2;
    int found = 0;
    int j = 0;
    BN_ULONG one = (unsigned long)0x8 << (BN_BITS2 - 4);


    /*Supress leading zeros to get an exact bit count*/
    for(j = 0; j < BN_BITS2; j++)
    {
        /*Decrement number of bits until first word is found*/
        if((!(a & one)) && (found != 1))
        {
            numberOfBits --;
            one >>= 1;
        }
        else
        {
            found = 1;
            break;
        }
    }
    

    return numberOfBits;

}
/*****************************************************************************************************/

    /*Input :  Big Number
      Output:  The number of bits in the number taking into account leading zeros
    */
    int countBits(BIGNUM * p)
    {
    
        int noOfBits = BN_BITS2 * p->top;
                                
        return noOfBits;
    }

/************************************************************************************************/
/*result, a and b must be different*/
int mult(BIGNUM *result, BIGNUM *a, BIGNUM *b)
{
    int i = 0;
    int j = 0;
    
    BN_ULONG c = 0;
    BN_ULONG* tempA;
    BN_ULONG* tempB;

    tempA = a->d;
    tempB = b->d;

    if(a == b)
    {
            square(result, a);
    }
    else
    {
        result->neg = a->neg | b->neg;

        bn_wexpand(result, a->top + b->top);

        for(i = 0; i < result->max; i++)
                result->d[i] = 0;
        result->top = a->top + b->top;
        for(i = 0; i < a->top; i++)
        {
            c = 0;
            for(j = 0; j < b->top; j++)
            {
                    mul_add(result->d[i + j], tempA[i], tempB[j], c);
            }

            result->d[i+j] = c;
    
        }

    }/*end else*/
    bn_fix_top(result);
    return 1;
}

/***********************************************************************************************************/
        
        /*
          This function uses Euclid's extended GCD algorithm to find r^-1 and n', two values 
          needed for calculating Montgomery multiplication and reduction. x in this case is the
          modulus and this is used to calulate n' and y is the number r which translates to 2^r mod n.
          This is used to find modInverse (r^-1).

          The algorithm for this is shown below.
  
        
          */
        
        int extendedGCD(BIGNUM* GCD, BIGNUM *modInverse, BIGNUM *nPrime, BIGNUM *x, BIGNUM *y, BN_CTX *ctx)
        {
                /*Algorithm
                 
                1.      g = 1
                2.      while ( y is even ) && ( x is even)
                                x = x / 2
                                y = y / 2
                                g = 2g
                3.  u = x
                        v = y
                        A = 1
                        B = 0
                        C = 0
                        D = 1

                4.      while (u is even)
                                u = u / 2
                                if ( A is congruent to B ie 0 mod 2)
                                        A = A / 2
                                        B = B / 2
                                else
                                        A = (A + y) / 2
                                        B = (B - x) / 2
                5.      while (v is even)
                                v = v / 2
                                if ( C is congruent to D ie 0 mod 2)
                                        C = C / 2
                                        D = D / 2
                                else
                                        C = (C + y) / 2
                                        D = (D - x) / 2

                6.      if u >= v
                                u = u - v
                                A = A - C
                                B = B - D

                        else
                                v = v - u
                                C = C - A
                                D = D - B

                7.      if u = 0
                                a = C
                                b = D
                        else
                                step 4

        */

        /*       u will never be greater in size than x and v will never be greater in size
                than y
        */
        /*      From the algorithm you can see that there are a lot of temporary variables used.
                As I mentioned before, I need to keep control over the number of mallocs and frees
                I use so again, a temporary variable structure is passed in which has be allocated
                already in main.  If someone wished to use this for other purposes, they can see
                from the algorithm above which temporary variable relates to which array position
                in the temporary struct and they can create their own temporary variables in this 
                function instead of sending in a pre-allocated struct if they wish
        */
        

                BIGNUM* u;
                BIGNUM* v;
                BIGNUM* A;
                BIGNUM* B;
                BIGNUM* C;
                BIGNUM* D;
                BIGNUM* g;
                BIGNUM* two;
                BIGNUM* tempX;
                BIGNUM* tempY;
                


                int i = 0;
                

                /*
                  If there are no more temporary variables available in ctx reinitialise it
                  to empty
                */
                if(ctx->tos == BN_CTX_NUM + 1)
                {
                        for(i = 0; i <= BN_CTX_NUM; i++)
                                BN_clear_free(&ctx->bn[i]);
                        BN_CTX_init(ctx);
                        ctx->flags=BN_FLG_MALLOCED;
                }

                
                u = &(ctx->bn[ctx->tos]);
                v = &(ctx->bn[ctx->tos+1]);
                A = &(ctx->bn[ctx->tos+2]);
                B = &(ctx->bn[ctx->tos+3]);
                C = &(ctx->bn[ctx->tos+4]);
                D = &(ctx->bn[ctx->tos+5]);
                g = &(ctx->bn[ctx->tos+6]);
                two = &(ctx->bn[ctx->tos+7]);
                tempX = &(ctx->bn[ctx->tos+8]);
                tempY = &(ctx->bn[ctx->tos+9]);

                ctx->tos += 10;
                
                BN_set_word(two, 2);
                

                BN_copy(tempX, x);
                BN_copy(tempY, y);
                
                /* g = 1*/
                BN_one(g);

                while((!BN_is_odd(tempX)) && (!BN_is_odd(tempY)))
                {
                        /* x = x / 2 */
                        BN_rshift(tempX, tempX, 1);
                        /* y = y / 2 */
                        BN_rshift(tempY, tempY, 1);
                        
                        /* g = 2g  - This is not needed so it is ignored here*/
                        mult(g, g, two);
                        
                }
                
                /* 
                u = x
                v = y
                A = 1
                B = 0
                C = 0
                D = 1
                */

                BN_copy(u, tempX);
                BN_copy(v, tempY);
                BN_one(A);
                BN_zero(B);
                BN_zero(C);
                BN_one(D);
                
        
                /*while u is not 0*/    
                while((!BN_is_word(u,0)) && (!BN_is_zero(u)))
                {
                        /*while u is even*/
                        while(!BN_is_odd(u))
                        {
                                /* u = u / 2*/
                                BN_rshift(u , u, 1);
                                /*if A and B are both even*/
                                if((!BN_is_odd(A)) && ( !BN_is_odd(B)))
                                {
                                        /*A = A / 2 */
                                        BN_rshift(A, A, 1);
                                        /*B = B / 2*/
                                        BN_rshift(B, B, 1);
                                
                                }
                                else
                                {
                                        /*A = (A + y) / 2*/
                                        BN_add(A,A,tempY);
                                        BN_rshift(A, A, 1);
                                
                                        /*B = (B - x) / 2*/
                                        BN_sub(B,B,tempX);
                                        BN_rshift(B, B, 1);
                                
                                }/*end if - else */
                        }/*end while */
        
                        /*while v is even*/
                        while(!BN_is_odd(v))
                        {
                                /* v = v / 2*/
                                BN_rshift(v, v, 1);
                                /*If C and D are even */
                                if((!BN_is_odd(C)) && ( !BN_is_odd(D)))
                                {
                                        /*C = C / 2*/
                                        BN_rshift(C, C, 1);
                                        /*D = D / 2 */
                                        BN_rshift(D, D, 1);
                                }
                                else
                                {
                                        /*C = (C + y) / 2*/
                                        BN_add(C,C,tempY);
                                        BN_rshift(C, C, 1);
                                        /*D = (D - x) / 2*/
                                        BN_sub(D, D,tempX);
                                        BN_rshift(D, D, 1);
                                }/*end if-else */
                        }/*end while*/
        

                        /*If u >= v
                           u = u - v
                           A = A - C
                           B = B - D
                        */
                        if(BN_ucmp(u, v) >= 0)
                        {
                                BN_sub(u, u, v);
                                BN_sub(A, A, C);
                                BN_sub(B, B, D);
                        }
                        /*v = v- u
                          C = C - A
                          D = D - B
                        */
                        else
                        {
                                BN_sub(v, v, u);
                                BN_sub(C, C, A);
                                BN_sub(D, D, B);
                        }/*end if - else*/

                        if((BN_is_word(u,0)) || (BN_is_zero(u)))
                        {
                                if((modInverse != NULL) && (nPrime != NULL))
                                {
                                        BN_copy(modInverse,D);
                                        BN_copy(nPrime,C);
                                        
                                }

                                if(GCD != NULL)
                                {       
        
                                        mult(GCD, g, v);
                                }
                                
                        }/*end if*/

                }/*end while u != 0 */
        
                        
                BN_clear_free(u);
                BN_clear_free(v);
                BN_clear_free(A);
                BN_clear_free(B);
                BN_clear_free(C);
                BN_clear_free(D);
                BN_clear_free(g);
                BN_clear_free(two);
                BN_clear_free(tempX);
                BN_clear_free(tempY);
                return 1;
        }
/**************************************************************************************************************/
        /*This function implements modular exponentiation using Montgomery methods,
          which allows you calculate the result without ever exceeding twice the modulus
          size for any calculation.

          This method will only work for odd moduli.
        */
        int modExp(BIGNUM * result, BIGNUM *a, BIGNUM *e, BIGNUM *n)
        {

                int i = 0;
                int noOfBitsInExponent = 0;
                int noOfBitsInModulus = 0;
                int currentPosition = 0;
                int counter = 0;
                BIGNUM* nPrime;
                BIGNUM* resOfA;
                BIGNUM* resOfX;
                BIGNUM* x;
                BIGNUM* R;
                BIGNUM* rInverse;
                BIGNUM* two;
                BIGNUM* temp;
                BIGNUM* RModN;
                BIGNUM* one;
                BN_CTX* divisionCTX;
                BN_CTX* monProCTX;
                BN_CTX* extendedGCDCTX;
                BN_CTX* squareCTX;
                BN_ULONG mask;
                                /*
                clock_t timeStart;
                clock_t timeEnd;
                                */
                int noOfBitsInA = 0;


                bn_fix_top(a);
                bn_fix_top(e);
                bn_fix_top(n);

                noOfBitsInA = countExactNumberOfBits(a);
                noOfBitsInExponent = countExactNumberOfBits(e);
                noOfBitsInModulus = countExactNumberOfBits(n);

                /*printf("\nThis calculation involves a(n) %d bit number raised to the power of a(n)\n%d bit number mod a(n) %d bit number\n",noOfBitsInA, noOfBitsInExponent, noOfBitsInModulus);
                */

                if(noOfBitsInExponent == 0)
                {
                        printf("\nWarning!  Attempt to raise something to the power of zero!\n");
                        BN_one(result);
                        return 1;
                }

                
                if(noOfBitsInA == 0)
                {
                        printf("\nWarning!  Zero Entry!\n");
                        BN_zero(result);
                        return 1;

                }
                
                /*timeStart = clock();*/

                /*Modulus must be odd*/
                if(!BN_is_odd(n))
                {
                        printf("\nError!  An Even Modulus has been Detected!  This compromises the security of the system!\nYou must use an odd modulus!\n");
                        return 0;
                }

                
                /*Initialise temporary variables which are to be used throughout this program*/
                extendedGCDCTX = BN_CTX_new();
                divisionCTX = BN_CTX_new();
                monProCTX = BN_CTX_new();
                squareCTX = BN_CTX_new();
                                
                /*
                  Allocate correct amount of space for each variable.  Doing this here ensures
                  that there are no unnecessary mallocs or frees which could result in 
                  degragmentation of memory later on in the program
                */
                

                if(noOfBitsInExponent == 1)
                {
                        printf("\nWarning!  Exponent = 1\n");
                        BN_mod(result, a, n, divisionCTX);
                        return 1;
                }

                if(BN_is_word(e, 2))
                {
                        rSquaredModN(result, a, n,divisionCTX);
                        return 1;
                }
                        
                

                /*Copy a into a temporary variable tempA and calcuate tempA mod n*/
                BN_mod(a, a, n, divisionCTX);
                

                /*If a mod n is 0 than the result is 0*/
                if((BN_is_word(a,0)) ||( BN_is_zero(a)))
                {

                        BN_zero(result);
                        return 1;
                }
                
                nPrime = BN_new();
                resOfA = BN_new();
                resOfX = BN_new();
                x = BN_new();
                R = BN_new();
                rInverse = BN_new();
                two = BN_new();
                temp = BN_new();
                RModN = BN_new();
                one = BN_new();
                                
                BN_init(nPrime);
                BN_init(resOfA);
                BN_init(resOfX);
                BN_init(x);
                BN_init(R);
                BN_init(rInverse);
                BN_init(two);
                BN_init(one);
                BN_init(temp);
                BN_init(RModN);
                

                BN_set_word(one,1);

                /*Allocate space for variables*/
                bn_wexpand(resOfX, n->top);
                bn_wexpand(rInverse, n->top * 2);
                bn_wexpand(nPrime, n->top * 2);
                bn_wexpand(temp, n->top);
                bn_wexpand(RModN, n->top);
                bn_wexpand(resOfA, a->top + n->top);
                
                
                /*R = 2 ^ noOfBitsInModulus*/
                calcR(R, noOfBitsInModulus);            
                
                calcRModN(RModN, R, n, divisionCTX);
                
                /*calculate n' and rInverse*/
                extendedGCD(NULL, rInverse, nPrime, n, RModN, extendedGCDCTX);

                /*calculate r Inverse mod n*/
                BN_mod(rInverse, rInverse, n, divisionCTX);
                
                
                /*Calucate the residue of a.
                  resA = a.r mod n
                */
                calcResidue(resOfA, a, RModN, n, divisionCTX);

                                
                /*Calulate the residue of x
                  resX = 1.r mod n
                */
                BN_copy(resOfX, RModN);
                
                                
                /*start at MSB*/
                currentPosition = e->top-1;
                counter = 0;
                mask = 0x80000000;
                while(!(mask & e->d[currentPosition]))
                {
                        mask >>= 1;
                        counter++;
                }
                
                
                /*For each bit in the exponent*/
                for(i = noOfBitsInExponent; i > 0; i--)
                {
                        
                        
                        if(counter >= 32)
                        {
                                currentPosition--;
                                mask = 0x80000000;
                                counter = 0;
                        }

                        /*Montgomery multiplication*/
                        MonPro(resOfX, resOfX, resOfX,  n, rInverse, monProCTX,  divisionCTX);

                        

                        /*If the current bit in the exponent is a one*/
                        if(e->d[currentPosition] & mask)
                        {
                                                                
                                MonPro(resOfX, resOfA, resOfX,  n, rInverse, monProCTX, divisionCTX );
                        }

                        mask >>= 1;
                        counter++;
                        
                }
        
                /*Convert the answer back to it's original form*/
                MonPro(x, resOfX, one, n, rInverse, monProCTX, divisionCTX);
                

                
                /*Ensure result is big enough to hold answer*/
                bn_wexpand(result, n->top);
                BN_copy(result, x);
                

                /*Free Big numbers*/
                BN_clear_free(nPrime);
                BN_clear_free(resOfA);
                BN_clear_free(resOfX);
                BN_clear_free(x);
                BN_clear_free(R);
                BN_clear_free(rInverse);
                BN_clear_free(two);
                BN_clear_free(temp);
                BN_clear_free(RModN);
                BN_clear_free(one);
                BN_CTX_free(divisionCTX);
                BN_CTX_free(extendedGCDCTX);
                BN_CTX_free(monProCTX);
                BN_CTX_free(squareCTX);
                                
                
                /*Calculate and display the length of time the operation took*/
                /*timeEnd = clock();
                printf( "%6.1f seconds\n", (double)(timeEnd 
                - timeStart)/CLOCKS_PER_SEC );
                printf("Running this test took %e seconds.\n", (double)(timeEnd 
                - timeStart)/CLOCKS_PER_SEC);
                
                */
                

                return 1;
        }
/**************************************************************************************************************/
        int MonPro(BIGNUM *result, BIGNUM *a, BIGNUM *b,  BIGNUM *modulus, BIGNUM* rInverseModN, BN_CTX* ctx, BN_CTX* divisionCTX)
        {

                BIGNUM* temp1;
                BIGNUM* temp2;
        
        
                int i = 0;

                if(ctx->tos == BN_CTX_NUM + 1)
                {
                
                        for(i = 0; i <= BN_CTX_NUM; i++)
                                BN_clear_free(&ctx->bn[i]);
                        BN_CTX_init(ctx);
                        ctx->flags=BN_FLG_MALLOCED;
                }

                temp1 = &(ctx->bn[ctx->tos]);
                temp2 = &(ctx->bn[ctx->tos+1]);

                ctx->tos += 2;
                
                if((a->top > modulus->top) || (b->top > modulus->top))
                        printf("\nError!\n");
                /* ab mod n*/
                mult(temp1, a, b);
                BN_mod(temp2, temp1, modulus, divisionCTX);
                        
                /* (ab mod n) rInverse mod n*/
                mult(temp1, temp2, rInverseModN);
                BN_mod(temp2, temp1, modulus, divisionCTX);
                
        
                if(BN_ucmp(temp2,modulus) >= 0)
                        BN_sub(result, temp2, modulus);
                else
                        BN_copy(result, temp2);


                BN_clear_free(temp1);
                BN_clear_free(temp2);
                return 1;

        }
/**************************************************************************************************************/        
        BYTE* bigNumberToBytes(BIGNUM *a, int *length)
        {

                BYTE* result;
                int noOfBytes;
                int noOfBits;
                BN_ULONG bit1 = 0x00000001;
                BN_ULONG bit2 = 0x00000002;
                BN_ULONG bit3 = 0x00000004;
                BN_ULONG bit4 = 0x00000008;
                BN_ULONG bit5 = 0x00000010;
                BN_ULONG bit6 = 0x00000020;
                BN_ULONG bit7 = 0x00000040;
                BN_ULONG bit8 = 0x00000080;
                int index = 0;
                int i = 0;
                int j = 0;
                int noOfShifts = 0;
                int bytesInOnePosition;

                if(BN_is_zero(a))
                {
                        result = (BYTE*)AEP_malloc (sizeof(BYTE)*1);
                        
                        mallocCount++;
                        result[0] = 0;
                        return result;
                }
                /*if there is one element with data, the number of bits is 1 * BN_BITS2*/
                noOfBits = a->top * BN_BITS2;
                noOfBytes = noOfBits/8;

                /*Each 32 bit chunk consists of 4 bytes*/
                bytesInOnePosition = BN_BITS2 / 8;

                /*allocate enough memory for the result*/
                result = (BYTE*)AEP_malloc(sizeof(BYTE)*noOfBytes);
                
                mallocCount++;
        
                for(i = 0; i < noOfBytes; i++)
                        result[i] = 0;

                /*LSB is in position 0 and MSB is in last position*/

                /*last position because want MSB in position 0 in order to successfully convert it to a
                  Java BigInteger
                */
                index = 0;

                for(i = a->top - 1; i >= 0; i--)
                {

                        noOfShifts = BN_BITS2 - 8;
                        j = 0;
                
                        /*go through four bytes*/
                        for(j = 0; j < bytesInOnePosition; j++)
                        {
                        
                                if(noOfShifts <= noOfBits)
                                {
                                        result[index] |= ((a->d[i] >> noOfShifts) & bit1);
                                        result[index] |= ((a->d[i] >> noOfShifts) & bit2);
                                        result[index] |= ((a->d[i] >> noOfShifts) & bit3);                                                              
                                        result[index] |= ((a->d[i] >> noOfShifts) & bit4);
                                        result[index] |= ((a->d[i] >> noOfShifts) & bit5);
                                        result[index] |= ((a->d[i] >> noOfShifts) & bit6);
                                        result[index] |= ((a->d[i] >> noOfShifts) & bit7);
                                        result[index] |= ((a->d[i] >> noOfShifts) & bit8);
                                        noOfShifts -= 8;
                                        index++;
                                }

                        }/*end for*/

                }/*end for*/


                *length = index;
                return result;
        }
/***********************************************************************************************************/
        
        void BN_CBytesToBigNum(BIGNUM* result, BYTE* number, int arrayLength)

        {
                int noOfPositions = 0;
                int i = 0;
                int j = 0;
                

                noOfPositions = arrayLength / BN_BYTES;

                if(arrayLength % BN_BYTES  > 0)
                        noOfPositions++;

                /*Set result top and initialise to 0*/
                bn_wexpand(result, noOfPositions);
                result->top = noOfPositions;
                for(i = 0; i < result->top; i++)
                        result->d[i] = 0;
                
                /*set the value of the BIGNUM from the bytes array*/
                /*Least significant byte is in position 0*/
        
                for(i = 0; i < noOfPositions; i++)
                {
                        result->d[i] = number[j];
                        result->d[i] |= (number[j+1] << 8);
                        result->d[i] |= (number[j+2] << 16);
                        result->d[i] |= (number[j+3] << 24);

                        j += 4;

                }
                                                                                     

                return;

        }
/***********************************************************************************************************/
        BIGNUM *bytesToBigNumber(BYTE* number, int length, int arrayLength)

        {
                float floatPositions;
                int intPositions;
                float difference;
                int noOfPositions = 0;
                BIGNUM * a;
                float temp1;
                float temp2;


                /*length indicates the number of bits in the Java Big Number*/
                temp1 = (float)length;
                /*indicates no. of bits supported by OS*/
                temp2 = (float)BN_BITS2;

                /*create and initialise a new big number*/
                a = BN_new();
                BN_init(a);
        
                /*if the number of bits is not a multiple of BN_BITS2 then assign an extra position*/
                floatPositions = (temp1 / temp2);
                intPositions = (length / BN_BITS2);


                difference = floatPositions - intPositions;

                /*allocate one extra position just in case*/
                if(difference > 0)
                        noOfPositions = intPositions + 1;
                else
                        noOfPositions = intPositions;

                /*set the value of the BIGNUM from the java BigInteger*/
                                                                                     /* no of bytes in array*/
                BN_SetNumberValue(a, number,arrayLength, noOfPositions);

        
                return a;

        }
/***********************************************************************************************************/
        /*a is the number to set
          number is the value to set it with
          length is the number of bytes in the Java BigInteger
          noOfPositions is the number of elements that will be needed in d[]
        */
        void BN_SetNumberValue(BIGNUM *a, BYTE* number, int length, int noOfPositions)
        {

                /*masks to check individual bits in a byte*/
                BN_ULONG bit1 = 0x00000001;
                BN_ULONG bit2 = 0x00000002;
                BN_ULONG bit3 = 0x00000004;
                BN_ULONG bit4 = 0x00000008;
                BN_ULONG bit5 = 0x00000010;
                BN_ULONG bit6 = 0x00000020;
                BN_ULONG bit7 = 0x00000040;
                BN_ULONG bit8 = 0x00000080;
        
                int index = 0;
                int i = 0;
                /*Number of bytes in each d[] position*/
                int noOfBytes = BN_BYTES;
                int j = 0;
                int noOfShifts = 0;

                bn_wexpand(a, noOfPositions);
                
                /*set the last index + 1*/
                a->top          = noOfPositions;
                
                /*Java BigInteger - MSB in number is in position 0 - LSB is in last
                  LSB should be in position 0 in this
                */
        
                /*MSB should be put in last position which is noOfPositions - 1*/

                /*last position*/
                index = length - 1;
                for(i = 0; i < noOfPositions; i++)
                {
        
                        noOfShifts = 0;
                        j = 0;
                
                        /*go through four bits*/
                        for(j = 0; j < noOfBytes; j++)
                        {       
                                if(index >= 0)
                                {
                                        /*initially noOfShifts = 0
                                         check if bit number 1 is set by ANDing the number with bit1.  OR the result
                                         with a->d[i] which will just set bit1 to the correct value without changing
                                         any of the other bits.
                                     The left shift allows you to repeat this loop for all 32 bits of the number
                                        */
                                        a->d[i] |= ((number[index] & bit1) <<noOfShifts);
                                        a->d[i] |= ((number[index] & bit2) <<noOfShifts);
                                        a->d[i] |= ((number[index] & bit3) <<noOfShifts);
                                        a->d[i] |= ((number[index] & bit4) <<noOfShifts);
                                        a->d[i] |= ((number[index] & bit5) <<noOfShifts);
                                        a->d[i] |= ((number[index] & bit6) <<noOfShifts);
                                        a->d[i] |= ((number[index] & bit7) <<noOfShifts);
                                        a->d[i] |= ((number[index] & bit8) <<noOfShifts);
                                                                        
                                        noOfShifts += 8;
                                
                                        index--;
                                }/*end if*/
        
                        }/*end for*/
        
                }/*end for*/
        
        }       
/***********************************************************************************************************/
        
        int BN_div(BIGNUM *dv, BIGNUM *rm, BIGNUM *num, BIGNUM *divisor,
             BN_CTX *ctx)
        {
                int norm_shift,i,j,loop;
                BIGNUM *tmp,wnum,*snum,*sdiv,*res;
                BN_ULONG *resp,*wnump;
                BN_ULONG d0,d1;
                int num_n,div_n;
                BN_ULONG t1l,t1h,t2l,t2h,t3l,t3h,ql,qh,t3t;
                BN_ULONG q,n0,n1;
                BN_ULONG l0;


                /*bn_check_top(num);
                  bn_check_top(divisor);
                */

                if (BN_is_zero(divisor))
                {
                        printf("\nError! Cannot Divide by Zero\n");
                        return(0);
                }

                if (BN_ucmp(num,divisor) < 0)
                {
                        if (rm != NULL)
                                { if (BN_copy(rm,num) == NULL) return(0); }
                        if (dv != NULL) BN_zero(dv);
                                return(1);
                }

                if(ctx->tos == BN_CTX_NUM + 1)
                {
                        for(i = 0; i <= BN_CTX_NUM; i++)
                                BN_clear_free(&ctx->bn[i]);
                        BN_CTX_init(ctx);
                        ctx->flags=BN_FLG_MALLOCED;
                }
                
                tmp= &(ctx->bn[ctx->tos]);
                tmp->neg=0;
                snum= &(ctx->bn[ctx->tos+1]);
                sdiv= &(ctx->bn[ctx->tos+2]);
                if (dv == NULL)
                        res= &(ctx->bn[ctx->tos+3]);
                else    res=dv;

                ctx->tos +=4;
                
                /* First we normalise the numbers */
                norm_shift=BN_BITS2-((countExactNumberOfBits(divisor))%BN_BITS2);
                BN_lshift(sdiv,divisor,norm_shift);
                sdiv->neg=0;
                norm_shift+=BN_BITS2;
                BN_lshift(snum,num,norm_shift);
                snum->neg=0;
                div_n=sdiv->top;
                num_n=snum->top;
                loop=num_n-div_n;

                /* Lets setup a 'window' into snum
                 * This is the part that corresponds to the current
                 * 'area' being divided */
                BN_init(&wnum);
                wnum.d=  &(snum->d[loop]);
                wnum.top= div_n;
                wnum.max= snum->max+1; /* a bit of a lie */
        
                /* Get the top 2 words of sdiv */
                /* i=sdiv->top; */
                d0=sdiv->d[div_n-1];
                d1=(div_n == 1)?0:sdiv->d[div_n-2];
        
                /* pointer to the 'top' of snum */
                wnump= &(snum->d[num_n-1]);
        
                /* Setup to 'res' */
                res->neg= (num->neg^divisor->neg);
                if (!bn_wexpand(res,(loop+1))) return 0;
                res->top=loop;
                resp= &(res->d[loop-1]);

                /* space for temp */
                if (!bn_wexpand(tmp,(div_n+1))) return 0;

                if (BN_ucmp(&wnum,sdiv) >= 0)
                {
                        if (!BN_usub(&wnum,&wnum,sdiv)) return 0;
                        *resp=1;
                        res->d[res->top-1]=1;
                }
                else
                        res->top--;
                resp--;

                for (i=0; i<loop-1; i++)
                {
                
                        wnum.d--; wnum.top++;
                        n0=wnump[0];
                        n1=wnump[-1];
                        if (n0 == d0)
                                q=BN_MASK2;
                        else
                                q=bn_div_words(n0,n1,d0);

                        {

                        t1h=n0;
                        t1l=n1;
                        for (;;)
                        {
                                t2l=LBITS(d1); t2h=HBITS(d1);
                                ql =LBITS(q);  qh =HBITS(q);
                                mul64(t2l,t2h,ql,qh); /* t2=(BN_ULLONG)d1*q; */

                                t3t=LBITS(d0); t3h=HBITS(d0);
                                mul64(t3t,t3h,ql,qh); /* t3=t1-(BN_ULLONG)q*d0; */
                                t3l=(t1l-t3t)&BN_MASK2;
                                if (t3l > t1l) t3h++;
                                t3h=(t1h-t3h)&BN_MASK2;


                        if (t3h) break;
                        if (t2h < t3l) break;
                        if ((t2h == t3l) && (t2l <= wnump[-2])) break;

                        q--;
                        }

                }
                l0=bn_mul_words(tmp->d,sdiv->d,div_n,q);
                tmp->d[div_n]=l0;
                for (j=div_n+1; j>0; j--)
                        if (tmp->d[j-1]) break;
                tmp->top=j;

                j=wnum.top;
                BN_sub(&wnum,&wnum,tmp);

                snum->top=snum->top+wnum.top-j;

                if (wnum.neg)
                        {
                        q--;
                        j=wnum.top;
                        BN_add(&wnum,&wnum,sdiv);
                        snum->top+=wnum.top-j;
                        }
                *(resp--)=q;
                wnump--;
                }
                
                if (rm != NULL)
                {
                        if(countExactNumberOfBits(snum) >= norm_shift)
                                BN_rshift(rm,snum,norm_shift);
                        else
                                BN_copy(rm, snum);
                        rm->neg=num->neg;
                
                }

                BN_clear_free(tmp);
                BN_clear_free(snum);
                BN_clear_free(sdiv);
                
                if(dv == NULL)
                        BN_clear_free(res);
                return(1);

        }

/***********************************************************************************************************/
        BN_ULONG bn_mul_words(BN_ULONG *rp, BN_ULONG *ap, int num, BN_ULONG w)
        {
                BN_ULONG c1=0;

                bn_check_num(num);
                if (num <= 0) return(c1);

                /* for (;;) */
                while (1) /* circumvent egcs-1.1.2 bug */
                {
                        mul(rp[0],ap[0],w,c1);
                        if (--num == 0) break;
                        mul(rp[1],ap[1],w,c1);
                        if (--num == 0) break;
                        mul(rp[2],ap[2],w,c1);
                        if (--num == 0) break;
                        mul(rp[3],ap[3],w,c1);
                        if (--num == 0) break;
                        ap+=4;
                        rp+=4;
                        }
                return(c1);
        } 
/***********************************************************************************************************/
        BN_ULONG bn_div_words(BN_ULONG h, BN_ULONG l, BN_ULONG d)
        {
                return((BN_ULONG)(((((BN_ULLONG)h)<<BN_BITS2)|l)/(BN_ULLONG)d));
        }

/***********************************************************************************************************/
        int BN_lshift(BIGNUM *r, BIGNUM *a, int n)
        {
                int i,nw,lb,rb;
                BN_ULONG *t,*f;
                BN_ULONG l;

                r->neg=a->neg;
                if (bn_wexpand(r,a->top+(n/BN_BITS2)+1) == NULL) return(0);
                nw=n/BN_BITS2;
                lb=n%BN_BITS2;
                rb=BN_BITS2-lb;
                f=a->d;
                t=r->d;
                t[a->top+nw]=0;
                if (lb == 0)
                        for (i=a->top-1; i>=0; i--)
                                t[nw+i]=f[i];
                else
                        for (i=a->top-1; i>=0; i--)
                        {
                                l=f[i];
                                t[nw+i+1]|=(l>>rb)&BN_MASK2;
                                t[nw+i]=(l<<lb)&BN_MASK2;
                        }
                memset(t,0,nw*sizeof(t[0]));
                
                r->top=a->top+nw+1;
                bn_fix_top(r);
                return(1);
        }
/***********************************************************************************************************/

        /* rem != m */
        int BN_mod(BIGNUM *rem, BIGNUM *m, BIGNUM *d, BN_CTX *ctx)
        {
                
                int dividendNegative = 0;

                if(m->neg == 1)
                {
                        m->neg = 0;
                        dividendNegative = 1;
                }
                
                BN_div(NULL,rem,m,d,ctx);

                /*If the dividend is negative*/
                if(dividendNegative == 1)
                {
                        m->neg = 1;
                        /*      If the remainder is smaller than the tempDivisor add the two
                                eg. -23 mod 25 = 2 mod 25 (25 + -23)
                        */
                        
                        if(BN_ucmp(rem,d) < 0)  
                        {
                                rem->neg = 1;
                                BN_add(rem,rem, d);
                                
                                
                        }
                        /*Otherwise subtract the remainder from the divisor to get the real
                          result
                        */
                        else
                        {
                                rem->neg = 1;
                                BN_usub(rem, d, rem);
                                                                
                        }


                }

                return 1;


        }
/***********************************************************************************************************/
        int expMinusOne(BIGNUM* result, BIGNUM *exp)
        {       
                int i = 0;
                int j = 0;
                
                BN_copy(result, exp);
                
                bn_fix_top(result);

                i = 0;
                while((result->d[i] == 0) && (i < result->top))
                        i++;

                result->d[i] -= 1;

                for(j = i-1; j >= 0; j--)
                        result->d[i] |= BN_MASK2;
                
                return 1;

        }
/***********************************************************************************************************/
        int mPlusOne(BIGNUM *result, BIGNUM *m)
        {
                
                int i = 0;
                int loop = 0;
                

                /*If LSB is FFFFFFFF - add one to next word and set following words to zero*/
                BN_copy(result, m);

                i = 0;

                while(result->d[i] & BN_MASK2)
                {
                        i++;
                        loop = 1;
                }

                if(i == result->top)
                {
                        bn_wexpand(result, result->top+1);
                }

                result->d[i] += 1;

                if(loop == 1)
                {
                        for(i = result->d[i-2]; i >= 0; i--)
                                result->d[i] = 0;
                }

                return 1;


        }
/***********************************************************************************************************/
        int mPlusOneOverTwo(BIGNUM *result, BIGNUM *m)
        {
                mPlusOne(result, m);

                BN_rshift(result, result, 1);

                return 1;

        }
/***********************************************************************************************************/

        int square(BIGNUM *result, BIGNUM *a)
        {
                int i = 0;
                
                BN_ULONG* tempA;
                BN_ULONG* tempB;
                
                
                tempA = a->d;
                tempB = (BN_ULONG*) AEP_malloc(sizeof(BN_ULONG) * (a->top * 2));
                result->neg = 0;

                bn_wexpand(result, a->top * 2);
                for(i = 0; i < result->max; i++)
                        result->d[i] = 0;

                result->top = a->top * 2;

                bn_sqr(result->d, tempA, a->top, tempB);
                
                bn_fix_top(result);
                                
                Free(tempB);
                
                return 1;
        }

/***********************************************************************************************************/
        BN_ULONG bn_mul_add_words(BN_ULONG *rp, BN_ULONG *ap, int num, BN_ULONG w)
        {
                BN_ULONG c1=0;

                bn_check_num(num);
                if (num <= 0) return(c1);

                for (;;)
                {
                        mul_add(rp[0],ap[0],w,c1);
                        if (--num == 0) break;
                        mul_add(rp[1],ap[1],w,c1);
                        if (--num == 0) break;
                        mul_add(rp[2],ap[2],w,c1);
                        if (--num == 0) break;
                        mul_add(rp[3],ap[3],w,c1);
                        if (--num == 0) break;
                        ap+=4;
                        rp+=4;
                }
        
                return(c1);
        } 
/***********************************************************************************************************/
        void bn_sqr_words(BN_ULONG *r, BN_ULONG *a, int n)
    {
                bn_check_num(n);
                if (n <= 0) return;
                for (;;)
                {
                        BN_ULLONG t;

                        t=(BN_ULLONG)(a[0])*(a[0]);
                        r[0]=Lw(t); r[1]=Hw(t);
                        if (--n == 0) break;

                        t=(BN_ULLONG)(a[1])*(a[1]);
                        r[2]=Lw(t); r[3]=Hw(t);
                        if (--n == 0) break;

                        t=(BN_ULLONG)(a[2])*(a[2]);
                        r[4]=Lw(t); r[5]=Hw(t);
                        if (--n == 0) break;

                        t=(BN_ULLONG)(a[3])*(a[3]);
                        r[6]=Lw(t); r[7]=Hw(t);
                        if (--n == 0) break;

                        a+=4;
                        r+=8;
                }
        }
/***********************************************************************************************************/
        void bn_sqr(BN_ULONG *r, BN_ULONG *a, int n, BN_ULONG *tmp)
        {
                int i,j,max;
                BN_ULONG *ap,*rp;

                max=n*2;
                ap=a;
                rp=r;
                rp[0]=rp[max-1]=0;
                rp++;
                j=n;

                if (--j > 0)
                {
                        ap++;
                        rp[j]=bn_mul_words(rp,ap,j,ap[-1]);
                        rp+=2;
                }

                for (i=n-2; i>0; i--)
                {
                        j--;
                        ap++;
                        rp[j]=bn_mul_add_words(rp,ap,j,ap[-1]);
                        rp+=2;
                }

                bn_add_words(r,r,r,max);
                
                /* There will not be a carry */

                bn_sqr_words(tmp,a,n);

                bn_add_words(r,r,tmp,max);
        }

/***********************************************************************************************************/
        int rSquaredModN(BIGNUM *result, BIGNUM *r, BIGNUM *n, BN_CTX* ctx)
        {
                
                square(result, r);
                BN_mod(result, result, n, ctx);

                return 1;
        }
/***********************************************************************************************************/
        int calcRModN(BIGNUM* RModN, BIGNUM* R, BIGNUM *n, BN_CTX* ctx)
        {

                /*Calculate R mod n if R is greater than n*/
                if(BN_ucmp(R,n) > 0)
                {       
                        BN_mod(RModN, R, n, ctx);
                }
                else
                        BN_copy(RModN, R);

                return 1;

        }
/***********************************************************************************************************/
        int calcR(BIGNUM* R, int noOfBitsInModulus)
        {
                
                BIGNUM* two;

                two = BN_new();
                BN_init(two);
                

                /*Calculate R ie 2 ^ r where r is the number of bits in the modulus*/
                BN_set_word(two, 2);
                BN_copy(R, two);
                
                /*Raise two to the power of noOfBitsInModulus.  It is quicker to shift left 
                  than to do an exponentiation operation
                */
                BN_lshift(R, R, noOfBitsInModulus-1);
                

                BN_clear_free(two);
                return 1;
        }
/***********************************************************************************************************/
        int calcResidue(BIGNUM *result, BIGNUM* a, BIGNUM* RModN, BIGNUM* n, BN_CTX* ctx)
        {

                /* a.R mod n*/
                mult(result, a, RModN);
                BN_mod(result, result, n, ctx);
                                        
                return 1;                       

        }
/***********************************************************************************************************/
/*These functions set the hardcoded numbers for the correctness tests*/
        BIGNUM* BN_SetCorrectnessBase()
        {
                BIGNUM* a;
                a = BN_new();
                BN_init(a);
                a->d = (BN_ULONG *)AEP_malloc(sizeof(BN_ULONG)*32); 
                
                a->top          = 32;
                a->max          = 32;           
                /*Each entry is 32 bits*/
                a->d[31] = 0x008c3e10uL;
a->d[30] = 0x9cce1c42uL;
a->d[29] = 0x8150c78duL;
a->d[28] = 0x42f12e10uL;
a->d[27] = 0x3a5a44fbuL;
a->d[26] = 0x259bd3dbuL;
a->d[25] = 0x60b32937uL;
a->d[24] = 0x9adb288buL;
a->d[23] = 0x15d2b323uL;
a->d[22] = 0x4950434auL;
a->d[21] = 0x8fbe1c0duL;
a->d[20] = 0xd28aaab8uL;
a->d[19] = 0xd90c7c77uL;
a->d[18] = 0x2b5a6267uL;
a->d[17] = 0x2920a56buL;
a->d[16] = 0xd875b0dbuL;
a->d[15] = 0x3b416fe3uL;
a->d[14] = 0xad5f5b89uL;
a->d[13] = 0xc8ba6e1duL;
a->d[12] = 0xb9fdd754uL;
a->d[11] = 0x7ca8cfb8uL;
a->d[10] = 0x4d9bcb62uL;
a->d[9] = 0x9a309ceeuL;
a->d[8] = 0x23fd7b7duL;
a->d[7] = 0xa76e2da0uL;
a->d[6] = 0xbedf524buL;
a->d[5] = 0xfebd774fuL;
a->d[4] = 0xafe3dccauL;
a->d[3] = 0xdd4a9f2duL;
a->d[2] = 0xc247ae06uL;
a->d[1] = 0xc2f32b64uL;
a->d[0] = 0xdcd4df4euL;
                
                return a;
        }
        
        BIGNUM* BN_SetCorrectnessPower()
        {
                BIGNUM* a;
                a = BN_new();
                BN_init(a);
                a->d = (BN_ULONG *)AEP_malloc(sizeof(BN_ULONG)*32); 
                
                a->top          = 32;
                a->max          = 32;           
                /*Each entry is 32 bits*/
                a->d[31] = 0x1049fcdauL;
a->d[30] = 0x1694e047uL;
a->d[29] = 0xe94116c7uL;
a->d[28] = 0x376e6e67uL;
a->d[27] = 0xd320e24buL;
a->d[26] = 0x51b885e4uL;
a->d[25] = 0xcc7f78f5uL;
a->d[24] = 0x11fac5a1uL;
a->d[23] = 0x96f6fa5buL;
a->d[22] = 0x44f0b94euL;
a->d[21] = 0x1c5d5733uL;
a->d[20] = 0x9717666cuL;
a->d[19] = 0xd24a046auL;
a->d[18] = 0xd75c6489uL;
a->d[17] = 0x88993fd3uL;
a->d[16] = 0x0d2f83ebuL;
a->d[15] = 0x4444c8a6uL;
a->d[14] = 0x9036c178uL;
a->d[13] = 0x5e18bbfeuL;
a->d[12] = 0xca2166a6uL;
a->d[11] = 0x16709366uL;
a->d[10] = 0x7ccf992duL;
a->d[9] = 0x1bcf6cfcuL;
a->d[8] = 0x55488472uL;
a->d[7] = 0x9b403545uL;
a->d[6] = 0x35b11d83uL;
a->d[5] = 0xeb5fdba9uL;
a->d[4] = 0x8e087390uL;
a->d[3] = 0x9ee90c19uL;
a->d[2] = 0xc15c28efuL;
a->d[1] = 0x80533b6fuL;
a->d[0] = 0x1e8fdc31uL;
                
                return a;
        }
        BIGNUM* BN_SetCorrectnessMod()
        {
                BIGNUM* a;
                a = BN_new();
                BN_init(a);
                a->d = (BN_ULONG *)AEP_malloc(sizeof(BN_ULONG)*32); 
                a->top          = 32;
                a->max          = 32;           
                /*Each entry is 32 bits*/
                a->d[31] = 0x008c3e10uL;
a->d[30] = 0x9cce1c42uL;
a->d[29] = 0x8150c78duL;
a->d[28] = 0x42f12e10uL;
a->d[27] = 0x3a5a44fbuL;
a->d[26] = 0x259bd3dbuL;
a->d[25] = 0x60b32937uL;
a->d[24] = 0x9adb288buL;
a->d[23] = 0x15d2b323uL;
a->d[22] = 0x4950434auL;
a->d[21] = 0x8fbe1c0duL;
a->d[20] = 0xd28aaab8uL;
a->d[19] = 0xd90c7c77uL;
a->d[18] = 0x2b5a6267uL;
a->d[17] = 0x2920a56buL;
a->d[16] = 0xd875b0dbuL;
a->d[15] = 0x3b416fe3uL;
a->d[14] = 0xad5f5b89uL;
a->d[13] = 0xc8ba6e1duL;
a->d[12] = 0xb9fdd754uL;
a->d[11] = 0x7ca8cfb8uL;
a->d[10] = 0x4d9bcb62uL;
a->d[9] = 0x9a309ceeuL;
a->d[8] = 0x23fd7b7duL;
a->d[7] = 0xa76e2da0uL;
a->d[6] = 0xbedf524buL;
a->d[5] = 0xfebd774fuL;
a->d[4] = 0xafe3dccauL;
a->d[3] = 0xdd4a9f2duL;
a->d[2] = 0xc247ae06uL;
a->d[1] = 0xc2f32b64uL;
a->d[0] = 0xdcd4df4fuL;
                
                return a;
        }
        
        BIGNUM* BN_SetCorrectnessP()
        {
                BIGNUM* a;
                a = BN_new();
                BN_init(a);
                a->d = (BN_ULONG *)AEP_malloc(sizeof(BN_ULONG)*16); 
                a->top          = 16;
                a->max          = 16;           
                /*Each entry is 32 bits*/
                a->d[15] = 0x0967f466uL;
a->d[14] = 0x57a97543uL;
a->d[13] = 0xe1be3904uL;
a->d[12] = 0x19808712uL;
a->d[11] = 0x4a4da7e5uL;
a->d[10] = 0xc35328afuL;
a->d[9] = 0xb0980eb6uL;
a->d[8] = 0xf4713ffauL;
a->d[7] = 0x8dd5af35uL;
a->d[6] = 0x8a6386bbuL;
a->d[5] = 0x84ed5e24uL;
a->d[4] = 0x4c15f9f0uL;
a->d[3] = 0xa61d8236uL;
a->d[2] = 0xb6142420uL;
a->d[1] = 0x78b66aa3uL;
a->d[0] = 0xa7649585uL;
                
                
                return a;
        }
        
        BIGNUM* BN_SetCorrectnessQ()
        {
                BIGNUM* a;
                a = BN_new();
                BN_init(a);
                a->d = (BN_ULONG *)AEP_malloc(sizeof(BN_ULONG)*16); 
                a->top          = 16;
                a->max          = 16;           
                /*Each entry is 32 bits*/
                a->d[15] = 0x0ee8e713uL;
a->d[14] = 0xbeee606buL;
a->d[13] = 0x81639be3uL;
a->d[12] = 0x4f0ea70fuL;
a->d[11] = 0x57d87898uL;
a->d[10] = 0xa4c802a8uL;
a->d[9] = 0x0b976828uL;
a->d[8] = 0x547a29f9uL;
a->d[7] = 0x594a1061uL;
a->d[6] = 0xbff58484uL;
a->d[5] = 0xa287f0aduL;
a->d[4] = 0x5a3fb0cbuL;
a->d[3] = 0xd7ad15ccuL;
a->d[2] = 0x915c73e1uL;
a->d[1] = 0x71cf36acuL;
a->d[0] = 0x42027fc3uL;
                
                return a;
        }
        BIGNUM* BN_SetCorrectnessDmpi()
        {
                BIGNUM* a;
                a = BN_new();
                BN_init(a);
                a->d = (BN_ULONG *)AEP_malloc(sizeof(BN_ULONG)*16); 
                
                a->top          = 16;
                a->max          = 16;           
                /*Each entry is 32 bits*/
                a->d[15] = 0x029d4c9cuL;
a->d[14] = 0x2ea22152uL;
a->d[13] = 0x0ac94a0duL;
a->d[12] = 0xf1508e51uL;
a->d[11] = 0xea994cf3uL;
a->d[10] = 0xfa4d9f8auL;
a->d[9] = 0xef90750auL;
a->d[8] = 0xf522e396uL;
a->d[7] = 0xeb02dd6fuL;
a->d[6] = 0x7f52300duL;
a->d[5] = 0x065779deuL;
a->d[4] = 0xb0ff4ad1uL;
a->d[3] = 0x4de775d3uL;
a->d[2] = 0xdad8ff8auL;
a->d[1] = 0xecccddf5uL;
a->d[0] = 0x8f1f4ffduL;
                
                
                return a;
        }
        BIGNUM* BN_SetCorrectnessDmqi()
        {
                BIGNUM* a;
                a = BN_new();
                BN_init(a);
                a->d = (BN_ULONG *)AEP_malloc(sizeof(BN_ULONG)*16); 
                
                a->top          = 16;
                a->max          = 16;           
                /*Each entry is 32 bits*/
                a->d[15] = 0x0cd77642uL;
a->d[14] = 0x45e7b157uL;
a->d[13] = 0x97a13fb2uL;
a->d[12] = 0x75218de3uL;
a->d[11] = 0xc139bf0euL;
a->d[10] = 0x86f15ddauL;
a->d[9] = 0xf140bfe8uL;
a->d[8] = 0xfe710cc7uL;
a->d[7] = 0xc7295adauL;
a->d[6] = 0x61636e2auL;
a->d[5] = 0x1cac3c1cuL;
a->d[4] = 0xba0e35c2uL;
a->d[3] = 0x1d8f0236uL;
a->d[2] = 0xf0ae2140uL;
a->d[1] = 0xce1ff3a5uL;
a->d[0] = 0x13baf147uL;
                
                
                return a;
        }
        BIGNUM* BN_SetCorrectnessIqmp()
        {
                BIGNUM* a;
                a = BN_new();
                BN_init(a);
                a->d = (BN_ULONG *)AEP_malloc(sizeof(BN_ULONG)*16); 
                
                a->top          = 16;
                a->max          = 16;           
                /*Each entry is 32 bits*/
                a->d[15] = 0x079a76efuL;
a->d[14] = 0x7e649f6buL;
a->d[13] = 0x9cfb70ceuL;
a->d[12] = 0x684a7dbbuL;
a->d[11] = 0x5af40affuL;
a->d[10] = 0x3278ce77uL;
a->d[9] = 0xce8beb82uL;
a->d[8] = 0xe1318498uL;
a->d[7] = 0x528c0bc5uL;
a->d[6] = 0x746abb27uL;
a->d[5] = 0x23a387e6uL;
a->d[4] = 0x3602f883uL;
a->d[3] = 0x467f4d7duL;
a->d[2] = 0xa962406euL;
a->d[1] = 0x42ac1e64uL;
a->d[0] = 0xc27f63dduL;
                
                return a;
        }
        
        BIGNUM* BN_SetCorrectnessResult()
        {
                BIGNUM* a;
                a = BN_new();
                BN_init(a);
                a->d = (BN_ULONG *)AEP_malloc(sizeof(BN_ULONG)*32); 
                a->top          = 32;
                a->max          = 32;   
                
                a->d[31] = 0x008c3e10uL;
                a->d[30] = 0x9cce1c42uL;
                a->d[29] = 0x8150c78duL;
                a->d[28] = 0x42f12e10uL;
                a->d[27] = 0x3a5a44fbuL;
                a->d[26] = 0x259bd3dbuL;
                a->d[25] = 0x60b32937uL;
                a->d[24] = 0x9adb288buL;
                a->d[23] = 0x15d2b323uL;
                a->d[22] = 0x4950434auL;
                a->d[21] = 0x8fbe1c0duL;
                a->d[20] = 0xd28aaab8uL;
                a->d[19] = 0xd90c7c77uL;
                a->d[18] = 0x2b5a6267uL;
                a->d[17] = 0x2920a56buL;
                a->d[16] = 0xd875b0dbuL;
                a->d[15] = 0x3b416fe3uL;
                a->d[14] = 0xad5f5b89uL;
                a->d[13] = 0xc8ba6e1duL;
                a->d[12] = 0xb9fdd754uL;
                a->d[11] = 0x7ca8cfb8uL;
                a->d[10] = 0x4d9bcb62uL;
                a->d[9] = 0x9a309ceeuL;
                a->d[8] = 0x23fd7b7duL;
                a->d[7] = 0xa76e2da0uL;
                a->d[6] = 0xbedf524buL;
                a->d[5] = 0xfebd774fuL;
                a->d[4] = 0xafe3dccauL;
                a->d[3] = 0xdd4a9f2duL;
                a->d[2] = 0xc247ae06uL;
                a->d[1] = 0xc2f32b64uL;
                a->d[0] = 0xdcd4df4euL;
                
                return a;
        }

