/*****************************************************************************
 *****************************************************************************
 Copyright (c) 2000-2001, AEP Ltd. 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.

 3. Neither the name of AEP Ltd. nor the names of its contributors 
 may be used to endorse or promote products derived from this software 
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

 *****************************************************************************
****************************************************************************/

/**********************************************************************
*                                                                     *
* AEP Ltd.                                                            *
*                                                                     *
* This software is supplied under the terms of the license included   *
* above.  All use of this software must be in accordance with the     *
* terms of that license.                                              *
*                                                                     *
**********************************************************************/

#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>

#include "aep_i_defs.h"
#include "aep_i_types.h"

#include "aependian.h"
#include "aepmalloc.h"
#include "aep_api.h"
#include "aep_api_globals.h"

#include "crc.h"
#include "load_utils.h"
#include "loader_version.h"

typedef struct AEP_BOOT_CMD
{
	AEP_U32 EntryAddress;
	AEP_U32 StartAddress;
	AEP_U32 TargetLength;
	AEP_U32 ExpectedCRC;
} AEP_BOOT_CMD;

typedef AEP_BOOT_CMD* AEP_BOOT_CMD_PTR;


#define AEP_BOOT_IMG_ADDRESS	0x00040000		/* Default address on target where target boot starts */
#define AEP_BLOCK_SIZE			0x00000800		/* 2 K size buffer block to read/write to target */
#define AEP_BOOT_CMD_SIZE		sizeof(AEP_BOOT_CMD) /* size in bytes of the message for booting
														the target (COMMAND ONLY, without header)*/
#define AEP_BOOT_MSG_SIZE AEP_BOOT_CMD_SIZE+HDR_SIZE+sizeof(AEP_U32) /* size in bytes of the FULL message
																		for booting the target (command + length + header) */
#define RESET_WAIT_TIME			2				/* Delay after reset command before downloading target */

/* Return values		*/
enum LoaderRVs {
	RV_NO_ERROR = 0,
	RV_APP_ALREADY_LOADED,
	RV_WRITE_ERROR,
	RV_READ_ERROR,
	RV_CRC_INCORRECT,
	RV_VERSION_DISPLAYED,
	RV_MISSING_ARGS,
	RV_DEVICE_ERR,
	RV_IMAGE_ERR,
	RV_TARGET_ERR,
	RV_UNKNOWN_ERR
} ;


/* Function Prototypes */
extern int BootTarget( int , FILE *, AEP_U32, AEP_U32 );
extern int ResetDev( int iDevFD );
extern int usage( void );


/************************************************/
/* Global Variables								*/
extern int bVerbose;
extern int bDoVersion;
extern int bDoPause;

extern int dev_fd;				/* AEP crypto accelerator file descriptor */
extern FILE *img_fd;

extern char acAEPDeviceName[100];	/* the name of the file associated with dev_fd, eg /dev/paep0 */
extern char acAEPImageName[100];	/* the name of the file with the target to load */


