/*****************************************************************************
 *****************************************************************************
 Copyright (c) 2000-2001, AEP Ltd. 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.

 3. Neither the name of AEP Ltd. nor the names of its contributors 
 may be used to endorse or promote products derived from this software 
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

 *****************************************************************************
****************************************************************************/

/**********************************************************************
*                                                                     *
* AEP Ltd.                                                            *
*                                                                     *
* This software is supplied under the terms of the license included   *
* above.  All use of this software must be in accordance with the     *
* terms of that license.                                              *
*                                                                     *
**********************************************************************/


/*This is a simple application to get the AEP version information. It simply makes an API call
  (AEP_GetVersion) and prints the results to standard output. The structure of the AEP version 
  information is defined in version.h. Specifying a -v command line option makes the program
  print it's own version number (i.e. the version number of the version application)*/

#include <stdio.h>
#include <string.h>
#include "aep_defs.h"
#include "aep_i_defs.h"
#include "aep_api.h"
#include "aep_version.h"
#include "versionapp_version.h"

#define VERSION_FLAG "-v"

int
main(int argc, char *argv[])
{
	AEP_RV rv;
	
	AEP_CONNECTION_HNDL firstConnection;
	
	int counter;
	
	AEP_CHAR versionString[100];
	
	/*Version struct*/
	AEP_VERSION versions;
	
	/*Whether it is being run with the version flag (-v) or not*/
	int versionFlag = 0;
	
	
	/*Check whether the -v option has been given on the command line*/
	for(counter = 1; counter < argc; counter ++)
	{
		if (strcmp (argv[counter], VERSION_FLAG)==0)
			versionFlag = 1;
		else
		{
			printf("Illegal option: %s \nCorrect usage: aepversion [-v]\n", argv[counter]);
			return (1);
		}
	}
	
	
	
	/*If the version flag was given, print a banner with the version number of this program*/
	if(versionFlag)
	{
		strcpy(versionString, AEPVERSION_VERSION );
		printf("AEP Version Application (version %s %s)\n", versionString, COMPILER_VERSION);
	} 
	
	else
	{
		/*Initialize the API & check for errors*/
		rv = AEP_Initialize(NULL);
		/*Check whether this was successful*/
		if(rv != AEP_R_OK)
		{
			printf("\nError initializing the API. ");
			switch(rv){
			case AEP_R_AEPAPI_ALREADY_INITIALIZED:
				printf("API is already initialized\n");
				break;
				
				/*Errors when trying to open a connection*/
			case AEP_R_NO_MORE_CONNECTION_HNDLS:
				printf("Unable to get connection handle\n");
				break;	
				/*Failure to connect to Daemon results in an AEP_R_GENERAL_ERROR*/
			case AEP_R_GENERAL_ERROR:
				printf("Could not connect to the daemon.\nIs the daemon running?\n");
				break;
				
				/*Error doing the genRandom in the AEP_Initialize call*/
			case AEP_R_HOST_MEMORY:
				printf("Host memory error");
				break;
			case AEP_R_SOCKERROR:
				printf("Socket error");
				break;
			case AEP_R_SOCKEOF:
				printf("Socket EOF");
				break;
			case AEP_R_ARGUMENTS_BAD:
				printf("Invalid arguments to message");
				break;
			case AEP_R_NO_TARGET_RESOURCES:
				printf("Insufficient resources on target to complete request");
				break;
			case AEP_R_TRANSACTION_HANDLE_INVALID:
				printf("Invalid transaction handle");
				break;
			case AEP_R_TIMED_OUT:
				printf("Timed out");
				break;
			case AEP_R_FXN_NOT_IMPLEMENTED:
				printf("Function not implemented");
				break;
			case AEP_R_TARGET_ERROR:
				printf("Target Error");
				break;
			case AEP_R_DAEMON_ERROR:
				printf("Daemon Error");
				break;
				
				/*Other error*/
			default:
				printf("Error code %x\n", rv);
			}/*End switch*/
			return(rv);
		}
		
		/*Open a connection to the daemon*/
		rv = AEP_OpenConnection(&firstConnection);
		if (rv != AEP_R_OK)
		{
			printf("Error! Could open a connection to the daemon. \n\n");
			
			/*Check all the different return values to see what error occurred*/
			if(rv == AEP_R_AEPAPI_NOT_INITIALIZED)
				printf("The API is not initialised.\n");
			/*If it is unable to get a connection handle it returns AEP_R_NO_MORE_CONNECTION_HANDLES*/
			else if (rv == AEP_R_NO_MORE_CONNECTION_HNDLS)
				printf("Unable to get connection handle\n");
			/*Failure to connect to Daemon results in an AEP_R_GENERAL_ERROR*/
			else if (rv == AEP_R_GENERAL_ERROR)
				printf("\n");
			
			return(rv);
		}  
		
		
		/*Make the API call to get the version information, and then print the results to the screen*/
		rv = AEP_GetVersion(firstConnection, &versions );
		if(rv != AEP_R_OK)
		{
			printf("Error getting version! ");
			switch(rv){
			case AEP_R_AEPAPI_NOT_INITIALIZED:
				printf("API not initialized\n");
				break;
			case AEP_R_HOST_MEMORY:
				printf("Host memory error\n");
				break;
			case AEP_R_SOCKERROR:
				printf("Socket error\n");
				break;
			case AEP_R_SOCKEOF:
				printf("Socket EOF\n");
				break;
			case AEP_R_ARGUMENTS_BAD:
				printf("Invalid arguments to message\n");
				break;
			case AEP_R_NO_TARGET_RESOURCES:
				printf("Insufficient resources on target to complete request\n");
				break;
			case AEP_R_TRANSACTION_HANDLE_INVALID:
				printf("Invalid transaction handle\n");
				break;
			case AEP_R_TIMED_OUT:
				printf("Timed out\n");
				break;
			case AEP_R_FXN_NOT_IMPLEMENTED:
				printf("Function not implemented\n");
				break;
			case AEP_R_TARGET_ERROR:
				printf("Target Error\n");
				break;
			case AEP_R_DAEMON_ERROR:
				printf("Daemon Error\n");
				break;
			default:
				printf("Error code: %0x \n", rv);
				break;
			}
			
			/*Close the connection and exit*/
			AEP_CloseConnection(firstConnection);
			return(rv);
		}

		printf("AEP API version: %s\n", versions.api_version.versionString);
		printf("AEP Daemon version: %s\n", versions.daemon_version.versionString);
		printf("Target version: %s\n", versions.target_version.target_version);
		printf("Target build date and time: %s\n", versions.target_version.target_build_date);
		printf("Target compiler version: %s\n", versions.target_version.compiler_version);
		printf("Micro engine build date and time: %s\n", versions.target_version.micro_engine_build_date);
		printf("CPLD version: %s\n", versions.target_version.cpld_version);
		printf("Board serial number and type: %s\n", versions.target_version.board_serial_no_and_type);
		printf("IXP1200 type: %s\n", versions.target_version.ixp1200_type);
		printf("Thermal RNG version: %s\n", versions.target_version.thermal_rng_version);
		printf("Thermal RNG build date: %s\n", versions.target_version.thermal_rng_build_date);
		printf("Thermal RNG build time: %s\n", versions.target_version.thermal_rng_build_time);
		
		/*Disconnect firstConnection*/
		rv = AEP_CloseConnection(firstConnection);
		if (rv != AEP_R_OK)
		{
			printf("Error! Could not disconnect from Daemon\n");
			
			if(rv == AEP_R_AEPAPI_NOT_INITIALIZED)
				printf("Error closing connection to the daemon: API not initialized.\n");
			
			return(rv);
		}  
		
		/*Clean up and exit*/ 
		rv = AEP_Finalize();
		if (rv != AEP_R_OK)
		{
			printf("Error finalizing\n");
			if(rv == AEP_R_AEPAPI_NOT_INITIALIZED)
				printf("Error finalizing: API not initialized\n");
			return(rv);
		} 
	}
		 
	return (AEP_R_OK);		
		
}


